
/********************************************************************************************************
 * mm use Header and footer for each block allocated/non allocated
 * footer helps to insert in free list in 0(1) time
 * also two-sided coalescing can be done in 0(1) time
 * freelist uses double-linked list
 * if either or both adjacent block are free it willbe merged with them
 * else block will be attached to head of list
 * free call will take o(1) time for coalescing andd insertion
 * MM_malloc first search for a block in free list if avilable using best first it will return
 * if not available new block will be requested from mem_sbrk
 
 *Metadata size in allocated block  is 8(footer)+8(header)=16 byte 
 * min block size will be !6 +8(min size after align)=24 byte
 *metadata size for free block will be 16(header)+8(footer)=24 byte
 *since min block size is guaranted to be 24 hence metadata for free block can be stored easily 
 *in giveen space
 
 ********************************************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>

#include "mm.h"
#include "memlib.h"

team_t team = {
    /* Team name */
    "Boolean Pundits",
    /* First member's full name */
    "paras garg",
    /* First member's email address */
    "parasgarg@cse.iitb.ac.in",
    /* Second member's full name (leave blank if none) */
    "Ketan kulkarni",
    /* Second member's email address (leave blank if none) */
    "ketankulkarni@cse.iitb.ac.in"
};

/* single word (4) or double word (8) alignment */
#define ALIGNMENT 8

/* rounds up to the nearest multiple of ALIGNMENT */
#define ALIGN(size) (((size) + (ALIGNMENT-1)) & ~0x7)


#define SIZE_T_SIZE (ALIGN(sizeof(size_t)))


#define ALLOC_UNIT sysconf(_SC_PAGESIZE)
//size of struct on 32bit system is 8 byte
typedef struct blk_header_t	{
	size_t size;
	int free;
	}blk_header;
//size of struct on 32bit system is 16 byte		
typedef struct free_blk_header_t	{
	size_t size;
	int free;
	struct free_blk_header_t *prev;
	struct free_blk_header_t *next;
	}free_blk_header;
//size of struct on 32 bit system is 8 byte	
typedef struct blk_footer_t	{
	int free;
	size_t size;
	}blk_footer;
   	
int METADATA_SIZE;

free_blk_header *head;
char* MAX_ADDRESS;
char* MIN_ADDRESS;
void *init_mem_sbrk_break = NULL;

//functions signature
void removeNodeFromFreeList(free_blk_header * bestfit);
void splitNode(free_blk_header *bestfit,int requiredSize);
void bothfree(blk_header* left,blk_header* curr,blk_header* righthead);
void bothNotFree(blk_header *header);
void leftFree(blk_header* left,blk_header* head);
void rightFree(blk_header* right,blk_header* currhead);
void printmem();
void printfreeList();

int mm_init(void)
{
	METADATA_SIZE= sizeof(blk_footer)+sizeof(blk_header);
	head=NULL;
	MIN_ADDRESS=mem_sbrk(0);
	//printf("ini%p",(void*)MIN_ADDRESS);
	//printf("starting address %p",(void*)MIN_ADDRESS);
	MAX_ADDRESS=MIN_ADDRESS;
	if(MIN_ADDRESS==NULL)
		return 0;

    return 1;		//Returns 0 on successfull initialization.
}

void removeNodeFromFreeList(free_blk_header * bestfit){
		if(bestfit->prev==NULL && bestfit->next==NULL){
			//printf("\n\n single node case bestfit %p",(void*)bestfit);
			head=NULL;
		//case b firs node
		}else if(bestfit->prev==NULL){
			free_blk_header *nextNode=bestfit->next;
			head=nextNode;
			nextNode->prev=NULL;
			//printf("\n\nfirst node %pnew head %p head prev%p head next%p",(void*)bestfit,(void*)head,(void*)head->prev,(void*)head->next);
		//case c last node
		}else if(bestfit->next==NULL){
			free_blk_header *prevNode=bestfit->prev;
			prevNode->next=NULL;
			//printf("\n\n last node %p prevnode %p",(void*)bestfit,(void*)prevNode);
		//case d middle node
		}else{
			free_blk_header *nextNode=bestfit->next;
			free_blk_header *prevNode=bestfit->prev;
			nextNode->prev=prevNode;
			prevNode->next=nextNode;
			//printf("\n\n middle node %p prevnode next %p nextnode prev%p",(void*)bestfit,(void*)prevNode->next,(void*)nextNode->prev);
		}

}
void splitNode(free_blk_header *bestfit,int requiredSize){
		int remainSize=bestfit->size-requiredSize;
		//SET BESTFIT SIZE TO REQUIRED SIZE
			
		free_blk_header *remainBlk=(free_blk_header*)((char*)bestfit+requiredSize);
		remainBlk->size=remainSize;
		remainBlk->free=1;
			
		blk_footer *remBlkFooter=(blk_footer*)((char*)remainBlk+remainSize-sizeof(blk_footer)); 
		remBlkFooter->size=remainSize;
		remBlkFooter->free=1;
			
		if(bestfit->next==NULL && bestfit->prev==NULL){//case single node
			remainBlk->next=NULL;
			remainBlk->prev=NULL;
			head=remainBlk;
		}else if(bestfit->next==NULL){//case last node
			bestfit->prev->next=remainBlk;
			remainBlk->next=NULL;
			remainBlk->prev=bestfit->prev;
		}else if(bestfit->prev==NULL){//case first node
			remainBlk->prev=NULL;
			remainBlk->next=bestfit->next;
			bestfit->next->prev=remainBlk;
			head=remainBlk;
			
		}else{//case middle node
			free_blk_header *prev=bestfit->prev;
			free_blk_header *next=bestfit->next;	
			prev->next=remainBlk;
			remainBlk->prev=prev;
			remainBlk->next=next;
			next->prev=remainBlk;
		}
}
//---------------------------------------------------------------------------------------------------------------
/* 
 * mm_malloc - Allocate a block by incrementing the brk pointer.
 *     Always allocate a block whose size is a multiple of the alignment.
 */
void *mm_malloc(size_t size)
{	
//printf("\n malloc %d",size);
	/* 
	 * This function should keep track of the allocated memory blocks.
	 * The block allocation should minimize the number of holes (chucks of unusable memory) in the heap memory.
	 * The previously freed memory blocks should be reused.
	 * If no appropriate free block is available then the increase the heap  size using 'mem_sbrk(size)'.
	 * Try to keep the heap size as small as possible.
	 */
	//printf("\n allocating %d",size);
	if(size <= 0){		// Invalid request size
		return NULL;
	}
	size=ALIGN(size);
	int requiredSize=size+METADATA_SIZE;
/*
 *search in free list if block not found besr
 */	
	
	free_blk_header *temp=head;
	free_blk_header *bestfit=NULL;
	//printmem();
	//printfreeList();
	while(temp!=NULL){
		//	printf("\n freelist search block address %p size %d",(void*)temp,temp->size);
		fflush(stdout);
		if(temp->size >= requiredSize){
			if(bestfit==NULL){
				bestfit=temp;
				}
			else if(bestfit->size>temp->size){
				bestfit=temp;
				}
		}
		temp=temp->next;	  
	}
	if(bestfit!=NULL){
/*	
 *IF BESTFIT SIZE GREATER THAN REQUIRED SIZE AND MIN BLKSIZE(METADAT+8BYTE) THEN SPLIT THE BLK AND ADJUST
 *POINTER TO REPLACE THE REMIAN BLK IN EXISTING FREE BLK LIST
  */	
		if(bestfit->size >= (requiredSize+8+METADATA_SIZE)){
			splitNode(bestfit,requiredSize);
		}
		else{
/*
 * since bestfit is of required size remove it form the freelist
 * change required size to full block size to allocate full block since blk cant be splitted in two halves
*/
			requiredSize=bestfit->size;
			removeNodeFromFreeList(bestfit);
		}
			
	}			
	blk_header* header=NULL;
	if(bestfit==NULL){
		header=(blk_header*)mem_sbrk(requiredSize);
		MAX_ADDRESS=(char*)header+requiredSize;
	}
	else{
		header=(blk_header*)bestfit;
	}
	
	header->size=requiredSize;
	header->free=0;
	
	//printf("\n size %d allocated %d",header->size,header->free);
	//set footer metadata
	blk_footer* footer=(blk_footer*)((char*)header+requiredSize-sizeof(blk_footer));
	//printf("\nheader %p size %d free %dfooter size %d free %d",(void*)header,header->size,header->free,footer->size,footer->free);
	footer->size=requiredSize;
	footer->free=0;
	
	//returing starting address for user data
	return (void*)((char*)header+sizeof(blk_header));	
	//mem_sbrk() is wrapper function for the sbrk() system call. 
	//Please use mem_sbrk() instead of sbrk() otherwise the evaluation results 
	//may give wrong results
}
void bothfree(blk_header* left,blk_header* curr,blk_header* righthead){

	free_blk_header *right=(free_blk_header*)righthead;
	//Arrange pointer for right side
	if(right->prev!=NULL && right->next!=NULL){
		free_blk_header *prev=right->prev;
		free_blk_header *next=right->next;
		//printf("\n\n prev prev%p  next%p",(void*)(prev->prev),(void*)(prev->next));
		//printf("\n\n next prev%p  next%p",(void*)(next->prev),(void*)(next->next));
		prev->next=next;
		next->prev=prev;
		//printf("\n\n prev prev%p  next%p",(void*)(prev->next),(void*)(next->prev));
		
		
	}else if(right->prev==NULL){
		//printf("\n\n\n prev head %p",(void*)head);
		head=right->next;
		right->next->prev=NULL;
		//printf("new head %p prev of head %p",(void*)head,(void*)head->prev);
	}else if(right->next==NULL){
		//printf("\n\n prev node %p prev new next%p",(void*)right->prev,(void*)right->prev->next);
		right->prev->next=NULL;
		//printf("\n\n prev node %p prev new next%p",(void*)right->prev,(void*)right->prev->next);
	}
		
	//now increase left size to leftsize+size+rightsize
	left->size=left->size+curr->size+right->size;
	blk_footer* footer=(blk_footer*)((char*)left+left->size-sizeof(blk_footer));
	footer->free=1;
	footer->size=left->size;
	//printf("\n\n left node left %p right %p",(void*)(((free_blk_header*)left)->prev),(void*)(((free_blk_header*)left)->next));
}
void bothNotFree(blk_header *header){
	int size=header->size;
	free_blk_header *free=(free_blk_header*)((void*)header);
	free->size=size;
	free->free=1;
	free->prev=NULL;
	free->next=NULL;
	//Attach block to start of list
	//printf("\n\n old head %p",(void*)head);
	if(head==NULL){
		head=free;
	}
	else{
		free->next=head;
		head->prev=free;
		head=free;
	}
	//printf("\nnew head %p prev %p next %p",(void*)head,(void*)head->prev,(void*)head->next);
}

void leftFree(blk_header* left,blk_header* head){
	//printf("\nleeft free");
	//simply increase block sizze no need to change the pointers
	//   free_blk_header *left1=(free_blk_header*)left;
	//printf("\nleft free %p prev %p next %p size %d free %d",(void*)left1,(void*)left1->prev,(void*)left1->next,left1->size,left1->free);
	int size=head->size;
	left->size=left->size+size;
	blk_footer* footer=(blk_footer*)((char*)left+left->size-sizeof(blk_footer));
	footer->free=1;
	footer->size=left->size;
	//printf("\nleft free %p prev %p next %p size %d free %d",(void*)left1,(void*)left1->prev,(void*)left1->next,left1->size,left1->free);
}

void rightFree(blk_header* right,blk_header* currhead){
	//printf("right free");
	int size=currhead->size;
	free_blk_header *prev=((free_blk_header*)right)->prev;
	free_blk_header *next=((free_blk_header*)right)->next;
	int rightsize=right->size;
	//set footer
	blk_footer* footer=(blk_footer*)((char*)right+right->size-sizeof(blk_footer));
	footer->size=rightsize+size;
	
	//set header
	free_blk_header *current=(free_blk_header*)currhead;
	current->free=1;
	current->size=rightsize+size;
	
	current->next=next;
	if(next!=NULL)
		next->prev=current;
	current->prev=prev;
	if(prev!=NULL){
		prev->next=current;
	}else{
	head=current;
	}
}
void mm_free(void *ptr)
{

	 if(ptr==NULL)
	 	return;
	 	
	 blk_header *header=(blk_header*)ptr-1;
	 //printf("\nfree %d",header->size);
	 int left_free=0;
	 int right_free=0;
	 
	 blk_header *rightHead=(blk_header*)((char*)header+header->size);
	 if((char*)rightHead!=MAX_ADDRESS){
	 //printf("right head exists");
	 	if(rightHead->free)
	 		right_free=1;
	 }
	 
	 blk_footer *prev_footer=(blk_footer*)header-1;
	 blk_header *leftHead=(blk_header*)((char*)header-prev_footer->size);
	// printf("prev head %p prev footer%p free %d next head%p free%d",(void*)leftHead,(void*)prev_footer,prev_footer->free,(void*)rightHead,rightHead->free);
	 if((char*)header!=MIN_ADDRESS){
	 //printf("left head exists");
	 	if(leftHead->free)
	 		left_free=1;
	 }
	// printf("\n right starting address  %p size %d free %d",(void*)rightHead,rightHead->size,rightHead->free);
	 //printf("\n left starting address  %p size %d free %d",(void*)leftHead,leftHead->size,leftHead->free);
	/*
	 * finding both adjacents blocks
	 * and check whether both are free
	*/
	if(left_free==1 && right_free==1){
		bothfree(leftHead,header,rightHead);
	}
	if(left_free==0 && right_free==1){
		rightFree(rightHead,header);
		}
	if(left_free==1 && right_free==0){
		leftFree(leftHead,header);
		}
		
	if(left_free==0 && right_free==0){
		bothNotFree(header);
	 }
}

/*
 * mm_realloc - Implemented simply in terms of mm_malloc and mm_free
 */
void *mm_realloc(void *ptr, size_t size)
{	
	if(ptr == NULL){			//memory was not previously allocated
		return mm_malloc(size);
	}
	
	if(size == 0){				//new size is zero
		mm_free(ptr);
		return NULL;
	}
	blk_header *curr=(blk_header*)((char*)ptr-sizeof(blk_header));
	int existingBlkSize=curr->size-METADATA_SIZE;
	if(size<=existingBlkSize)
		return ptr;

	free_blk_header *right=(free_blk_header*)((char*)curr+curr->size);
	if((char*)right<MAX_ADDRESS){
		if(right->free && (existingBlkSize+right->size)>=size){
			//printf("\n cru free %d newsixe %d re sxer%d",right->free,existingBlkSize+right->size,size);
			removeNodeFromFreeList(right);
			curr->size=curr->size+right->size;
			curr->free=0;
			blk_footer* footer=(blk_footer*)((char*)curr+curr->size-sizeof(blk_footer));
			footer->size=curr->size;
			footer->free=0;
			return ptr;
		}		
	}
	blk_footer *prev_footer=(blk_footer*)curr-1;
	 if((char*)curr!=MIN_ADDRESS){
	 	if(prev_footer->free&& (existingBlkSize+prev_footer->size)>=size){
	 		free_blk_header *left=(free_blk_header*)((char*)curr-prev_footer->size);
			removeNodeFromFreeList(left);
			left->size=curr->size+left->size;
			left->free=0;
			blk_footer* footer=(blk_footer*)((char*)left+left->size-sizeof(blk_footer));
			footer->size=left->size;
			footer->free=0;
			memcpy( (void*)((char*)left+sizeof(blk_header)),ptr,existingBlkSize);
			return (void*)((char*)left+sizeof(blk_header));
	 	}
	 }

	int bytesToCopy=existingBlkSize;
	if(existingBlkSize>size)
		bytesToCopy=size;
		
	void* memory=mm_malloc(size);
	memcpy(memory,ptr,bytesToCopy);
	mm_free(ptr);
	return memory;
	
}
void printmem(){
	blk_header *header;
	char* temp=MIN_ADDRESS;
	printf("\n printing mem %p",(void*)MIN_ADDRESS);
	fflush(stdout);
	int i=1;
	while(temp<MAX_ADDRESS){
		header=(blk_header*)((void*)temp);
		temp=temp+header->size;
		printf("\n Block %d Size %d free %d %p alinged %d ",i++,header->size,header->free,(void*)header,((intptr_t)(header+1)) % 8); 
	
	}
}

void printfreeList(){
	free_blk_header *header=head;
	printf("\n printing freelist mem %p",(void*)head);
	fflush(stdout);
	int i=1;
	while(header!=NULL){
		printf("\n free Block %d Size %d free %d %p ",i++,header->size,header->free,(void*)header); 
		header=header->next;
	}
}

