#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

#include "mm.h"
#include "memlib.h"

team_t team = {
    /* Team name */
    "Boolean Pundits",
    /* First member's full name */
    "paras garg",
    /* First member's email address */
    "parasgarg@cse.iitb.ac.in",
    /* Second member's full name (leave blank if none) */
    "Ketan kulkarni",
    /* Second member's email address (leave blank if none) */
    "ketankulkarni@cse.iitb.ac.in"
};

/* single word (4) or double word (8) alignment */
#define ALIGNMENT 8
/* rounds up to the nearest multiple of ALIGNMENT */
#define ALIGN(size) (((size) + (ALIGNMENT-1)) & ~0x7)
#define SIZE_T_SIZE (ALIGN(sizeof(size_t)))
#define HEIGHT(node) (node==NULL?0:node->height)

typedef struct blk_header_t	{
	int free;
	int size;
	}blk_header;
//size of struct on 32bit system is 20 byte		
typedef struct free_blk_header_t	{
	int free;
	int size;
	struct free_blk_header_t *prev;
	struct free_blk_header_t *next;
	}free_blk_header;
//size of struct on 32 bit system is 8 byte	
typedef struct blk_footer_t	{
	int free;
	size_t size;
	}blk_footer;

typedef struct seg_t{
	free_blk_header* head;
}seg;
seg *segList;
#define N_SEG_LIST 10
int METADATA_SIZE;
int MIN_BLOCK_SIZE;
free_blk_header *root=NULL;

char* MAX_ADDRESS;
char* MIN_ADDRESS;	

//function signature
void insertNodeAtFirst(free_blk_header *curr,int bin);
void removeNodeFromSegList(free_blk_header *node,int bin);
int getBin(int i);
void splitAndInsertRemainAtHead(free_blk_header *bestfit,int requiredSize);
void printmem();
free_blk_header* getBestFit(int requiredSize);

int mm_init(void)
{
	segList=mem_sbrk(N_SEG_LIST*sizeof(seg));
	for(int i=0;i<N_SEG_LIST;i++)
		segList[i].head=NULL;
	for(int i=0;i<N_SEG_LIST;i++)
		//printf("%p",segList[i].head);
	METADATA_SIZE=sizeof(blk_header)+sizeof(blk_footer);
	MIN_BLOCK_SIZE=METADATA_SIZE+8;
	MIN_ADDRESS=mem_sbrk(0);
	MAX_ADDRESS=MIN_ADDRESS;
	if(MIN_ADDRESS==NULL)
		return 1;
    return 0;		
}

int getBin(int i){
	if(i==24)
	return 0;
	int l=0;	
	while (i >>= 1) { ++l; }
	if(l>N_SEG_LIST+4)
		return 9; 
	return l-5;
}


void insertNodeAtFirst(free_blk_header *curr,int bin){
	curr->prev=NULL;
	curr->free=1;
	free_blk_header *head=segList[bin].head;
	
	if(head!=NULL){
		curr->next=head;
		head->prev=curr;
		
	}else{
		curr->next=NULL;
		}
	segList[bin].head=curr;
}

void removeNodeFromSegList(free_blk_header * node,int bin){
		if(node->prev==NULL && node->next==NULL){
			//printf("\n\n single node case bestfit %p",(void*)bestfit);
			segList[bin].head=NULL;
		//case b firs node
		}else if(node->prev==NULL){
			free_blk_header *nextNode=node->next;
			segList[bin].head=nextNode;
			nextNode->prev=NULL;
			//printf("\n\nfirst node %pnew head %p head prev%p head next%p",(void*)bestfit,(void*)head,(void*)head->prev,(void*)head->next);
		//case c last node
		}else if(node->next==NULL){
			free_blk_header *prevNode=node->prev;
			prevNode->next=NULL;
			//printf("\n\n last node %p prevnode %p",(void*)bestfit,(void*)prevNode);
		//case d middle node
		}else{
			free_blk_header *nextNode=node->next;
			free_blk_header *prevNode=node->prev;
			nextNode->prev=prevNode;
			prevNode->next=nextNode;
			//printf("\n\n middle node %p prevnode next %p nextnode prev%p",(void*)bestfit,(void*)prevNode->next,(void*)nextNode->prev);
		}

}

void splitAndInsertRemainAtHead(free_blk_header *bestfit,int requiredSize){
	int remainSize=bestfit->size-requiredSize;
	bestfit->size=requiredSize;
	
	free_blk_header *remain=(free_blk_header*)((char*)bestfit+requiredSize);
	remain->size=remainSize;
	blk_footer* footer=(blk_footer*)((char*)remain+remain->size-sizeof(blk_footer));
	footer->size=remain->size;
	footer->free=1;
	insertNodeAtFirst(remain,getBin(remainSize));
}

free_blk_header* getBestFit(int requiredSize){
	int bin=getBin(requiredSize);
	free_blk_header *list=segList[bin].head;
	free_blk_header *bestfit=NULL;
	//printf("\nmalloc reg bin %d",bin);
	while(list!=NULL){
		if(list->size>=requiredSize){
			bestfit=list;
			break;
			//code to find bestfit
			/*if(bestfit!=NULL){
				if(list->size<bestfit->size)
					bestfit=list;
			}else{
				bestfit=list;
			}*/
		}
		list=list->next;
	}
	
	if(bestfit!=NULL){
		removeNodeFromSegList(bestfit,bin);
		return bestfit;
	}
/*
 *Since,bestfit/firstfit is not found in given list go for next higher fit and split node if new node size>required size since requiredSize=METADATA_SIZE+ALIGN(Size) .so.2 times of it can hold 2 METADATA and 2 ALIGN(SiZE)
 */
	bin++;
	while(bin<N_SEG_LIST){
		if(segList[bin].head!=NULL){
			bestfit=segList[bin].head;
			break;	
		}
		bin++;
	}
	
	if(bestfit!=NULL){
		removeNodeFromSegList(bestfit,bin);
		if( bestfit->size >= 2*requiredSize){
			///printf("\nreq size%d new size%d",requiredSize,bestfit->size);
			splitAndInsertRemainAtHead(bestfit,requiredSize);
		}
	}
	return bestfit;
}

void *mm_malloc(size_t size)
{	
	//printmem();
	if(size <= 0){		// Invalid request size
		return NULL;
	}
	
	int requiredSize=ALIGN(size)+METADATA_SIZE;
	free_blk_header *bestfit=getBestFit(requiredSize);
	if(bestfit!=NULL){
	//to avoid internal memory leaking beacause due to align and avoided splitiing block may contain extra splace that leads to imternal fragmentation but that is manageable.
		requiredSize=bestfit->size;
	}
	
	blk_header *header=(blk_header*)bestfit;
	//printf("\n block header %p",header);
	if(header==NULL){
		header=(blk_header*)mem_sbrk(requiredSize);
		//set uper limit of allocated address
		MAX_ADDRESS=(char*)header+requiredSize;
	}
	//printf("\n block header %p",header);
	//printf("\n req size %d",requiredSize);
	header->size=requiredSize;
	header->free=0;
	blk_footer* footer=(blk_footer*)((char*)header+requiredSize-sizeof(blk_footer));
	footer->size=requiredSize;
	footer->free=0;
	//printf("\nheader %p size %d free %dfooter size %d free %d",(void*)header,header->size,header->free,footer->size,footer->free);	
	//returing starting address for user data
	return (void*)((char*)header+sizeof(blk_header));
}

void mm_free(void *ptr)
{
	
	 if(ptr==NULL)
	 	return;
	free_blk_header *curr=(free_blk_header*)((char*)ptr-sizeof(blk_header));
	free_blk_header *right=(free_blk_header*)((char*)ptr-sizeof(blk_header)+curr->size);
	int rightfree=0;
	if((char*)right<MAX_ADDRESS){
		if(right->free)
			rightfree=1;
	}
	int leftfree=0;
	free_blk_header *left=NULL;
	if((char*)curr!=MIN_ADDRESS){
		blk_footer *leftFooter=(blk_footer*)((char*)curr-sizeof(blk_footer));
		if(leftFooter->free){
			leftfree=1;
			left=(free_blk_header*)((char*)curr-leftFooter->size);
		}
	}
	if(leftfree && rightfree){
		int leftbin=getBin(left->size);
		int rightbin=getBin(right->size);
		removeNodeFromSegList(left,leftbin);
		removeNodeFromSegList(right,rightbin);
		left->size=curr->size+left->size+right->size;
		blk_footer* footer=(blk_footer*)((char*)left+left->size-sizeof(blk_footer));
		footer->size=left->size;
		footer->free=1;
		insertNodeAtFirst(left,getBin(left->size));
		
	}
	else if(leftfree){
		int leftbin=getBin(left->size);
		removeNodeFromSegList(left,leftbin);
		left->size=curr->size+left->size;
		
		blk_footer* footer=(blk_footer*)((char*)left+left->size-sizeof(blk_footer));
		footer->size=left->size;
		footer->free=1;
		insertNodeAtFirst(left,getBin(left->size));
		
	}
	else if(rightfree){
		int rightbin=getBin(right->size);
		removeNodeFromSegList(right,rightbin);
		curr->size=curr->size+right->size;
		
		blk_footer* footer=(blk_footer*)((char*)curr+curr->size-sizeof(blk_footer));
		footer->size=curr->size;
		footer->free=1;
		insertNodeAtFirst(curr,getBin(curr->size));
		}
	else{
		int bin=getBin(curr->size);
		
		blk_footer* footer=(blk_footer*)((char*)curr+curr->size-sizeof(blk_footer));
		footer->size=curr->size;
		footer->free=1;
		insertNodeAtFirst(curr,bin);
	}
	/* 
	 * Searches the previously allocated node for memory block with base address ptr.
	 * 
	 * It should also perform coalesceing on both ends i.e. if the consecutive memory blocks are 
	 * free(not allocated) then they should be combined into a single block.
	 * 
	 * It should also keep track of all the free memory blocks.
	 * If the freed block is at the end of the heap then you can also decrease the heap size 
	 * using 'mem_sbrk(-size)'.
	 */
}

void *mm_realloc(void *ptr, size_t size)
{	
	if(ptr == NULL){			//memory was not previously allocated
		return mm_malloc(size);
	}
	
	if(size == 0){				//new size is zero
		mm_free(ptr);
		return NULL;
	}
	
	blk_header *curr=(blk_header*)((char*)ptr-sizeof(blk_header));
	int existingBlkSize=curr->size-METADATA_SIZE;
	if(size<=existingBlkSize)
		return ptr;

	/*free_blk_header *right=(free_blk_header*)((char*)curr+curr->size);
	if((char*)right<MAX_ADDRESS){
		if(right->free && (existingBlkSize+right->size-METADATA_SIZE)>=size){
			//printf("\n cru free %d newsixe %d re sxer%d",right->free,existingBlkSize+right->size,size);
			int rightbin=getBin(right->size);
			removeNodeFromSegList(right,rightbin);
			curr->size=curr->size+right->size;
			curr->free=0;
			blk_footer* footer=(blk_footer*)((char*)curr+curr->size-sizeof(blk_footer));
			footer->size=curr->size;
			footer->free=0;
			return ptr;
		}		
	}*/
	 
	if((char*)curr!=MIN_ADDRESS){
		blk_footer *leftFooter=(blk_footer*)((char*)curr-sizeof(blk_footer));
		free_blk_header *left=(free_blk_header*)((char*)curr-leftFooter->size);
		if(leftFooter->free && (existingBlkSize+left->size-METADATA_SIZE)>=size){
			
			int leftbin=getBin(left->size);
			removeNodeFromSegList(left,leftbin);
			left->size=curr->size+left->size;
			left->free=0;
			blk_footer* footer=(blk_footer*)((char*)left+left->size-sizeof(blk_footer));
			footer->size=left->size;
			footer->free=0;
			memcpy( (void*)((char*)left+sizeof(blk_header)),ptr,existingBlkSize);
			return  (void*)((char*)left+sizeof(blk_header));		
		}
	}
	int bytesToCopy=existingBlkSize;
	if(existingBlkSize>size)
		bytesToCopy=size;
			
	void* memory=mm_malloc(size);
	memcpy(memory,ptr,bytesToCopy);
	mm_free(ptr);
	return memory;

}

void printmem(){
	blk_header *header;
	char* temp=MIN_ADDRESS;
	printf("\n printing mem %p",(void*)MIN_ADDRESS);
	fflush(stdout);
	int i=1;
		header=(blk_header*)((void*)temp);
	while(temp<MAX_ADDRESS){
		temp=temp+header->size;
		printf("\n Block %d Size %d free %d %p alinged %d ",i++,header->size,header->free,(void*)header,((intptr_t)(header+1)) % 8); 
	
	}
}
